#include "fat.internals.h"

#if FAT_WRITE
/**
 * Write buffered block with FAT table data back to memory
 * (FAT table uses multiple copies)
 */
bool _Fat_WriteFATBlock(const fat_state_t* pFS, sd_cache_t* pBuf) {
	uint16_t ret;
	uint32_t fatSize = pFS->FATSz;
	int i = 0;
	
	fatSize <<= pFS->BytsPerSec_bit;
	Sd_set_blocklength(9);
	do {
		ret = Sd_write_block( pBuf->address, pBuf->buffer);
		RETF(ret);
		i++;
		if( i >= pFS->NumFATs ) {
			pBuf->state &= ~SD_CACHE_DIRTY;
			SD_FREE_LOCK(pBuf);
			return true;
		}
		pBuf->address += fatSize;
	} while(1);
}
bool Fat_EraseCluster(fat_state_t* pFS, fat_ca_t cluster) {
	uint8_t	BlPerClus;
	uint16_t	ret;
	uint32_t	address;
	
	_Fat_getClusterPos( pFS, &cluster, &address );
	BlPerClus = 1 << (pFS->BytsPerClus_bit - 9);
	Sd_set_blocklength(9);
	do {
		ret = Sd_set_block( address, 0 );
		BlPerClus--;
	} while( BlPerClus );
	
	return( ret > 0 );
}

bool Fat_EraseChain(fat_state_t* pFS, fat_ca_t firstCA) {
	fat_ca_t	cluster = firstCA;
	bool		ret;
	
	do {
		ret = Fat_EraseCluster( pFS, cluster );
		RETF(ret);
	} while(_fat_getNextCA( pFS, cluster, &cluster ));
	
	return true;
}

/**
 * Create a chain of count clusters starting and return first of these.
 * @param[in/out]	pFirstCA	Pointer to buffer for number of first cluster of chain
 * @param[in]		count		Length of chain to create
 * @return			True if successful
 */
bool _Fat_CreateChain(fat_state_t* pFS, fat_ca_t* pFirstCA, uint16_t count) {
	fat_ca_t		lastCA = pFS->TotSec << pFS->SecPerClus_bit;
	fat_ca_t		prevCA = *pFirstCA;
	uint8_t			bl;
	uint32_t			blockAdr;
	uint16_t			entriesPerBlock;
	uint16_t			i;
	fat_ca_t		thisCA = pFS->LastFreeCA;
	sd_cache_t*		pBuf;
	
	#if FAT12_SUPPORT
		if( pFS->FATType & FAT12 )
			return false;
	#endif
	
	bl = FAT_ENTRY_WIDTH(pFS);
	
	blockAdr = pFS->FatSA;
	blockAdr <<= pFS->BytsPerSec_bit;
	blockAdr += pFS->LastFreeCA >> (9 - bl);
	entriesPerBlock = SD_WRITE_BLOCKLENGTH >> bl;
	thisCA = pFS->LastFreeCA;
	*pFirstCA = 0;
	do {
		pBuf = Sd_cache_read_block(&blockAdr);
		RETF(pBuf);

		for( i = 0; i < entriesPerBlock; i++, thisCA++ ) {
			if( ((FAT16_SEC*)pBuf->buffer)[i] == 0 ) {
				// return first new claimed cluster
				if( *pFirstCA == 0 )
					*pFirstCA = thisCA;
				// link last cluster
				if( prevCA ) {
					if( prevCA < pFS->LastFreeCA ) {
						SD_FREE_LOCK(pBuf);
						_Fat_Update(pFS, prevCA, thisCA);
						pBuf = Sd_cache_read_block(&blockAdr);
					} else {
						prevCA -= pFS->LastFreeCA;
						((FAT16_SEC*)pBuf->buffer)[prevCA] = thisCA;
						pBuf->state |= SD_CACHE_DIRTY;
					}
				}
				count--;
				if( count ) {
					prevCA = thisCA;
				} else {
					// end of chain
					i <<= bl;
					bl <<= 1;
					memset(&pBuf->buffer[i], 0xFF, bl);
					_Fat_WriteFATBlock(pFS, pBuf);
					return true;
				}
			}
		}

		_Fat_WriteFATBlock(pFS, pBuf);
		blockAdr += SD_WRITE_BLOCKLENGTH;
		pFS->LastFreeCA = thisCA;
	} while( pFS->LastFreeCA < lastCA );
	
	return false;
}

/**
 * Update FAT table entry cluster to point to nextCluster
 * \note Does not support 12 bit FAT
 * \todo FAT32 support
 */
bool _Fat_Update(fat_state_t* pFS, fat_ca_t cluster, fat_ca_t nextCluster) {
	uint32_t		address;
	uint16_t		ret;
	
	_Fat_getClusterFAT( pFS, &cluster, &address );
	ret = Sd_write(address, nextCluster, sizeof(fat_ca_t));
	return( ret != 0 );
}

bool _Fat_FileAppendCluster(fat_state_t* pFS, file_handle_t* hFile, uint16_t count) {
	if( _Fat_CreateChain(pFS, &hFile->CurrentCA, count) ) {
		hFile->Flags &= ~FILE_HANDLE_FLAG_EOF;
		hFile->ClusterOffset = 0;
		return true;
	}
	return false;
}

bool Fat_FileResize(fat_state_t* pFS, file_handle_t* hFile, unsigned long size) {
	fat_ca_t		newNumClus;
	fat_ca_t		curNumClus;
	uint32_t			address;
	bool			ret;
		
	curNumClus = hFile->Size >> (pFS->BytsPerSec_bit + pFS->SecPerClus_bit);
	newNumClus = size >> (pFS->BytsPerSec_bit + pFS->SecPerClus_bit);
		
	if( newNumClus > curNumClus ) {
		address = hFile->Size - hFile->Position;
		File_Seek( pFS, hFile, address, NULL );
		newNumClus-= curNumClus;
		_Fat_FileAppendCluster(pFS, hFile, curNumClus);
		hFile->Position += 1;
		
		Fat_EraseChain( pFS, hFile->CurrentCA );
	} else {
		File_Reset(hFile);
		// TODO shorten cluster chain!
	}

	//
	// Update dir entry
	//
	address = hFile->DirEntryPos + 28;
	ret = Sd_write(address, &size, sizeof(uint32_t));
	
	return ret != 0;
}

/**
 * Update contents an existing directory entry.
 */
bool Fat_UpdateDirEntry(
	fat_state_t*		pFS,
	uint32_t*				pAddress,
	struct fat_direntry*		pDirEntry
) {
	uint16_t	ret;
	
	//blockOffset = _Fat_AlignAddress(&blockAdr);
	ret = Sd_write(*pAddress, pDirEntry, sizeof(struct fat_direntry) );
	return( ret != 0 );
}

bool _Fat_CreateDirEntry(
	fat_state_t*	pFS,
	file_handle_t*	hParent,
	struct fat_direntry*	pDirEntry,
	uint32_t*			pAddress
) {
	struct fat_direntry*	pDE;
	uint16_t			ret;
	sd_cache_t*		pBuf;
	int 			i;
	
	File_Reset(hParent);
	do {
		*pAddress = _Fat_GetFilePos(pFS, hParent);
		
		pBuf = Sd_cache_read_block(pAddress);
		RETF(pBuf);
		pDE = (struct fat_direntry*)pBuf->buffer;
		
		for( i = 0; i < 16; i++, pDE++ )  {
			if( (pDE->Name[0] == (char)FAT_NAME_FLAG_DELETED) ||
				(pDE->Name[0] == (char)FAT_NAME_FLAG_END) )
			{
				memcpy( pDE, pDirEntry, sizeof(struct fat_direntry) );
				pBuf->state |= SD_CACHE_DIRTY;

				if( pDE->Name[0] == (char)FAT_NAME_FLAG_END ) {
					if( i == 15 ) {
						// worst case: new block needed
						SD_FREE_LOCK(pBuf);			
						File_Seek( pFS, hParent, SD_WRITE_BLOCKLENGTH, NULL );
						
						if( hParent->Flags & FILE_HANDLE_FLAG_EOF ) {
							if( hParent->StartCA == 1 ) {
								// root dir full
								return false;							
							} else {
								ret = _Fat_FileAppendCluster(pFS, hParent, 1);
								RETF(ret);
								*pAddress = _Fat_GetFilePos(pFS, hParent);
							}
						} else {
							// directory has more space (eg. clustersize > sectorsize)
							*pAddress += SD_WRITE_BLOCKLENGTH;
						}
						_Sd_cache_flush();
						SD_GET_LOCK(pBuf);
						pBuf->address = *pAddress;
						
						// clear new directory block
						memset(pBuf->buffer, 0, SD_WRITE_BLOCKLENGTH);
						pDE = (struct fat_direntry*)pBuf->buffer;
					} else {
						pDE++;
					}
					*((uint8_t*)pDE) = FAT_NAME_FLAG_END;
					pBuf->state |= SD_CACHE_DIRTY;
				}
				goto _Fat_CreateDirEntry_Done;
			}
		}
		
		SD_FREE_LOCK(pBuf);
		File_Seek( pFS, hParent, SD_WRITE_BLOCKLENGTH, NULL );
		if( hParent->Flags & FILE_HANDLE_FLAG_EOF ) {
			ret = _Fat_FileAppendCluster(pFS, hParent, 1);
			RETF(ret);
		}
	} while(1);
	
 _Fat_CreateDirEntry_Done:
	SD_FREE_LOCK(pBuf);
	Sd_flush();
	return true;
}

bool Fat_CreateFile(
	fat_state_t*		pFS,
	file_handle_t*		hParent,
	const char*			name,
	const uint8_t			attributes,
	const fat_date_t	date,
	const fat_time_t	time,
	const uint32_t		size,
	file_handle_t*		hFile,
	struct fat_direntry*		pDirEntry
) {
	struct fat_direntry	de;
	fat_ca_t		numC;
	uint8_t			ret;
	uint16_t			bps = (1 << pFS->BytsPerClus_bit) - 1;

	numC = size >> pFS->BytsPerClus_bit;
	if( size & bps )
		numC++;

	File_Close(hFile);
	ret = _Fat_CreateChain(pFS, &hFile->StartCA, numC);
	RETF(ret);		// disk full? create chain should be undone in case of failure
	
	//
	// Create directory entry and handle
	//
	memset(((uint8_t*)&de)+12, 0, sizeof(struct fat_direntry) - 12);
	Fat_CanonicalNameToFAT(de.Name, name);
	memcpy(&de.CrtDate, &date, 2);
	memcpy(&de.CrtTime, &time, 2);
	memcpy(&de.WrtDate, &date, 2);
	memcpy(&de.WrtTime, &time, 2);
	memcpy(&de.LstAccDate, &date, 2);
	memcpy(&de.FstClusLO, &hFile->StartCA, 2);
	de.Attributes = attributes;
		
	if( attributes & FAT_ATTR_DIRECTORY ) {
		hFile->Flags = FILE_HANDLE_FLAG_DIRECTORY;
	} else {
		memcpy(&de.FileSize, &size, 4);
	}
	hFile->Size = size;
	File_Reset(hFile);
	
	if( pDirEntry )
		memcpy(pDirEntry, &de, sizeof(struct fat_direntry));
	
	ret = _Fat_CreateDirEntry(pFS, hParent, &de, &hFile->DirEntryPos);
	return ret;
}

#endif
