/*
Copyright 2006, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2006
*/

#include "ScatterWeb.MD4.h"
#include "ScatterWeb.MD4.internals.h"
#include <string.h>

void MD4_memcpy(uint8_t* output, uint8_t* input, uint16_t len) {
  memcpy(output, input, len);
}

// interesting enough: the algorithm runs faster, when memset() is
// called indirectly through MD4_memset() - when calling memset()
// directly from the program, you lose about 0.05-0.1ms per hash
void MD4_memset(uint8_t* output, uint16_t value, uint16_t len) {
  memset(output, value, len);
}

// MD4 initialization. Begins an MD4 operation, writing a new context
void MD4_init(MD4_CTX *context) {
  context->count[0] = context->count[1] = 0;
  // Load magic initialization constants.
  context->state[0] = 0x67452301;
  context->state[1] = 0xefcdab89;
  context->state[2] = 0x98badcfe;
  context->state[3] = 0x10325476;
}

// MD4 basic transformation. Transforms state based on block
void MD4_transform(uint32_t *state, const uint8_t *block) {
  uint32_t a = state[0], b = state[1], c = state[2], d = state[3];
  uint32_t* x = (uint32_t*) block;
  // Round 1
  FF (a, b, c, d, x[ 0], S11); // 1
  FF (d, a, b, c, x[ 1], S12); // 2
  FF (c, d, a, b, x[ 2], S13); // 3
  FF (b, c, d, a, x[ 3], S14); // 4
  FF (a, b, c, d, x[ 4], S11); // 5
  FF (d, a, b, c, x[ 5], S12); // 6
  FF (c, d, a, b, x[ 6], S13); // 7
  FF (b, c, d, a, x[ 7], S14); // 8
  FF (a, b, c, d, x[ 8], S11); // 9
  FF (d, a, b, c, x[ 9], S12); // 10
  FF (c, d, a, b, x[10], S13); // 11
  FF (b, c, d, a, x[11], S14); // 12
  FF (a, b, c, d, x[12], S11); // 13
  FF (d, a, b, c, x[13], S12); // 14
  FF (c, d, a, b, x[14], S13); // 15
  FF (b, c, d, a, x[15], S14); // 16
  // Round 2
  GG (a, b, c, d, x[ 0], S21); // 17
  GG (d, a, b, c, x[ 4], S22); // 18
  GG (c, d, a, b, x[ 8], S23); // 19
  GG (b, c, d, a, x[12], S24); // 20
  GG (a, b, c, d, x[ 1], S21); // 21
  GG (d, a, b, c, x[ 5], S22); // 22
  GG (c, d, a, b, x[ 9], S23); // 23 
  GG (b, c, d, a, x[13], S24); // 24 
  GG (a, b, c, d, x[ 2], S21); // 25
  GG (d, a, b, c, x[ 6], S22); // 26 
  GG (c, d, a, b, x[10], S23); // 27
  GG (b, c, d, a, x[14], S24); // 28
  GG (a, b, c, d, x[ 3], S21); // 29
  GG (d, a, b, c, x[ 7], S22); // 30
  GG (c, d, a, b, x[11], S23); // 31
  GG (b, c, d, a, x[15], S24); // 32 
  // Round 3
  HH (a, b, c, d, x[ 0], S31); // 33
  HH (d, a, b, c, x[ 8], S32); // 34
  HH (c, d, a, b, x[ 4], S33); // 35
  HH (b, c, d, a, x[12], S34); // 36
  HH (a, b, c, d, x[ 2], S31); // 37
  HH (d, a, b, c, x[10], S32); // 38
  HH (c, d, a, b, x[ 6], S33); // 39
  HH (b, c, d, a, x[14], S34); // 40
  HH (a, b, c, d, x[ 1], S31); // 41
  HH (d, a, b, c, x[ 9], S32); // 42
  HH (c, d, a, b, x[ 5], S33); // 43
  HH (b, c, d, a, x[13], S34); // 44
  HH (a, b, c, d, x[ 3], S31); // 45
  HH (d, a, b, c, x[11], S32); // 46
  HH (c, d, a, b, x[ 7], S33); // 47
  HH (b, c, d, a, x[15], S34); // 48
  state[0] += a;
  state[1] += b;
  state[2] += c;
  state[3] += d;
}

// MD4 block update operation. Continues an MD4 message-digest
// operation, processing another message block, and updating the
// context
void MD4_update(MD4_CTX *context, const uint8_t *input, uint16_t inputLen) {
  uint16_t i, index, partLen;
  // Compute number of bytes mod 64
  index = (uint16_t)((context->count[0] >> 3) & 0x3F);
  // Update number of bits
  if ((context->count[0] += ((uint32_t)inputLen << 3)) < ((uint32_t)inputLen << 3)) {
    context->count[1]++;
  }
  context->count[1] += ((uint32_t)inputLen >> 29);
  partLen = 64 - index;
  // Transform as many times as possible
  if (inputLen >= partLen) {
    MD4_memcpy((uint8_t*) &context->buffer[index], (uint8_t*) input, partLen);
    MD4_transform(context->state, context->buffer);
    for (i = partLen; i + 63 < inputLen; i += 64) {
      MD4_transform (context->state, &input[i]);
    }
    index = 0;
  }
  else {
    i = 0;
  }
  // Buffer remaining input
  MD4_memcpy((uint8_t*) &context->buffer[index], (uint8_t*) &input[i], inputLen-i);
}

// MD4 finalization. Ends an MD4 message-digest operation, writing the
// the message digest and zeroizing the context.
void MD4_final(uint8_t *digest, MD4_CTX *context) {
  uint8_t bits[8];
  uint16_t index, padLen;
  // Save number of bits
  MD4_memcpy(bits, (uint8_t*) context->count, 8);
  // Pad out to 56 mod 64
  index = (uint16_t)((context->count[0] >> 3) & 0x3f);
  padLen = (index < 56) ? (56 - index) : (120 - index);
  MD4_update(context, PADDING, padLen);
  // Append length (before padding)
  MD4_update(context, bits, 8);
  // Store state in digest
  MD4_memcpy(digest, (uint8_t*) context->state, 16);
}

uint8_t* MD4_get(uint8_t *input, uint16_t length, uint8_t *output) {
  MD4_CTX context;
  MD4_init(&context);
  MD4_update(&context, input, length);
  MD4_final(output, &context);
  return output;
}
